/*
 * out_yasapi.c
 *
 * Yet Another (WA)SAPI (YASAPI) NT Output Plugin for Winamp (out_yasapi-nt)
 * Copyright (C) 2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <windows.h>
#include <shlobj.h>
#include <yasapi.h>
#if defined (YASAPI_OPTION_GAPLESS) // [
#include <winamp/wa_ipc.h>
#endif // ]
#if defined (YASAPI_API) // [
#define YASAPI_EXPORT
#include <yasapi_nt_api.h>
#endif // ]

#define YASAPI_ENGINE

#define PI_VER2 "v1.11 beta"

#ifdef __alpha
#define PI_VER PI_VER2 " (AXP)"
#else
#define PI_VER PI_VER2 " (x86)"
#endif

#if defined (YASAPI_OPTION_GAPLESS) // [
static int SubclassWinamp(HWND hWnd);
static int UnSubclassWinamp(HWND hWnd);
#endif // ]

int getwrittentime();
int getoutputtime();

int srate, numchan, bps, active;
volatile int writtentime, w_offset;

static Player player;


BOOL WINAPI _DllMainCRTStartup(HANDLE hInst, ULONG ul_reason_for_call, LPVOID lpReserved)
{
	return TRUE;
}

Out_Module out;

#if ! defined (YASAPI_ENGINE) // [
static int last_pause=0;
#endif // ]

#define YASAI_ON_TOP_OF_PREFERENCES
#if ! defined (YASAI_ON_TOP_OF_PREFERENCES) // [
#define YASAI_ON_TOP_OF_WINAMP_MAIN
#endif // ]
void config(HWND hwnd)
{
#if defined (YASAPI_ENGINE) // [
  PlayerShow(&player,&player.gtk.config,hwnd);
#endif // ]
}

void about(HWND hwnd)
{
#if defined (YASAPI_ENGINE) // [
  PlayerShow(&player,&player.gtk.about,hwnd);
#endif // ]
}

void init()
{
	PlayerCreate(&player,&out);
#if defined (YASAPI_OPTION_GAPLESS) // [
  if (SubclassWinamp(out.hMainWindow)<0) {
    DMESSAGE("subclassing winamp");
    goto subclass;
  }
subclass: ;
#endif // ]
}

void quit()
{
#if defined (YASAPI_OPTION_GAPLESS) // [
  UnSubclassWinamp(out.hMainWindow);
#endif // ]
	PlayerDestroy(&player);
}

int start_t;

int open(int samplerate, int numchannels, int bitspersamp, int bufferlenms, int prebufferms)
{
#if defined (YASAPI_ENGINE) && defined (YASAPI_OPTION_GAPLESS) // [
  int latency;
#endif // ]
  DMARKLN();
	start_t=GetTickCount();
 	w_offset = writtentime = 0;
	active=1;
	numchan = numchannels;
	srate = samplerate;
	bps = bitspersamp;
#if ! defined (YASAPI_ENGINE) // [
	return 1;
#elif defined (YASAPI_OPTION_GAPLESS) // ] [
  if (PlayerOpen(&player,samplerate,numchannels,bitspersamp,&latency,NULL)<0)
    return -1;
  else
    return latency;
#else // ] [
  return PlayerOpen(&player,samplerate,numchannels,bitspersamp);
#endif // ]
}

void close()
{
#if defined (YASAPI_ENGINE) // [
  DMARKLN();
  EngineClose(&player.engine);
  DMARKLN();
#endif // ]
}

int write(char *buf, int len)
{
#if defined (YASAPI_ENGINE) // [
  // 0 on success. Len == bytes to write (<= 8192 always). buf is straight
  //   audio data. 
  // 1 returns not able to write (yet). Non-blocking, always.
  if (PlayerWrite(&player,(BYTE *)buf,len)<0)
    return 1;
  else {
#endif // ]
	writtentime += len;
	return 0;
#if defined (YASAPI_ENGINE) // [
  }
#endif // ]
}

int canwrite()
{
#if defined (YASAPI_ENGINE) // [
  return PlayerAvailable(&player);
#else // ] [
	if (last_pause) return 0;
	if (getwrittentime() < getoutputtime()+MulDiv(65536,1000,srate*bps*numchan/8)) return 65536;
	return 0;
#endif // ]
}

int isplaying()
{
#if defined (YASAPI_ENGINE) // [
  return PlayerIsPlaying(&player);
#else // ] [
	return 0;
#endif // ]
}

#if defined (YASAPI_ENGINE) // [
int pause(int pause)
{
#if defined (YASAPI_DEBUG) // [
  static int count;
#endif // ]
  int p;
  
  //DMARKLN();
  DWRITELN("================================================================");
  DVWRITELN("count %d",++count);
  DWRITELN("================================================================");
  p=PlayerPause(&player,pause);
  //DVWRITELN("%d",p);

  return p;
}
#else // ] [
int pause(int pause)
{
	int t=last_pause;
	if (!last_pause && pause) { w_offset+=GetTickCount()-start_t;  writtentime=0; }
	if (last_pause && !pause) { start_t=GetTickCount(); }
	last_pause=pause;
	return t;
}
#endif // ]

void setvolume(int volume)
{
#if defined (YASAPI_ENGINE) // [
  PlayerVolume(&player,volume);
#endif // ]
}

void setpan(int pan)
{
}

void flush(int t)
{
#if defined (YASAPI_ENGINE) // [
  PlayerFlush(&player,t);
#else // ] [
  w_offset=t;
  start_t=GetTickCount();
  writtentime=0;
#endif // ]
}
	
int getoutputtime()
{
#if defined (YASAPI_ENGINE) // [
  return PlayerOutputTime(&player);
#else // ] [
	if (last_pause)
		return w_offset;
	return GetTickCount()-start_t + w_offset;
#endif // ]
}

int getwrittentime()
{
#if defined (YASAPI_ENGINE) // [
  return PlayerWrittenTime(&player);
#else // ] [
	int t=srate*numchan,l;
	int ms=writtentime;

	if (t)
	{
	l=ms%t;
	ms /= t;
	ms *= 1000;
	ms += (l*1000)/t;

	ms/=(bps/8);

	return ms + w_offset;
	}
	else
		return ms;
#endif // ]
}

#if defined (YASAPI_BETA) // [
#define OUT_VER_U 0x11
#endif // ]
Out_Module out = {
#if defined (OUT_VER_U) // [
	OUT_VER_U,
  (char *)YASAPI_TITLE_U16
#else // ] [
	OUT_VER,
  YASAPI_TITLE
#endif // ]
	,
	65,
	0, // hmainwindow
	0, // hdllinstance
	config,
	about,
	init,
	quit,
	open,
	close,
	write,
	canwrite,
	isplaying,
	pause,
	setvolume,
	setpan,
	flush,
	getoutputtime,
	getwrittentime
};

__declspec( dllexport ) Out_Module * winampGetOutModule()
{
	return &out;
}

#if defined (YASAPI_OPTION_GAPLESS) // [
#if 0 // [
// Would have been nice when working ...
#define OUT_UNSET		0x0		// no longer the active output
#define OUT_SET			0x1		// is becoming the active output
#define OUT_PLAYBACK	0x100	// used when playback begins and an output plug-in
								// is set for an input plug-in or not based on the
								// input plug-in settings when the playback begins

__declspec(dllexport) void __cdecl winampGetOutModeChange(int mode)
{
#if 0 // [
  if (OUT_UNSET==mode) {
DMESSAGE("HERE WE GO ...");
    PlayerClearActive(&player);
  }
#else // ] [
DVMESSAGE("mode: %03x",mode);
#endif // ]
}
#else // ] [
static WNDPROC WinampProc;

static LRESULT CALLBACK PluginWinampProc(HWND hWnd, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  char *modulea;

  switch (uMsg) {
  case WM_WA_IPC:
    switch (lParam) {
    case IPC_CB_OUTPUTCHANGED:
      modulea=(char *)SendMessageW(hWnd,WM_WA_IPC,0,IPC_GETOUTPUTPLUGIN);
      DVWRITELN("cmp: \"%s\" (%d)",modulea,yasapi_wcsstrcmp(player.pBasename,
          modulea));
      PlayerSetActive(&player,!yasapi_wcsstrcmp(player.pBasename,modulea));
      break;
    default:
      break;
    }

    break;
  default:
    break;
  }

  return CallWindowProc(
    WinampProc,   // _In_ WNDPROC lpPrevWndFunc,
    hWnd,         // _In_ HWND    hWnd,
    uMsg,         // _In_ UINT    Msg,
    wParam,       // _In_ WPARAM  wParam,
    lParam        // _In_ LPARAM  lParam
  );
}

static int SubclassWinamp(HWND hWnd)
{
  if (NULL==hWnd||NULL!=WinampProc) {
    DMESSAGE("failing to subclass winamp window");
    return -1;
  }

  WinampProc=(WNDPROC)SetWindowLong(
    hWnd,                   // _In_ HWND hWnd,
    GWL_WNDPROC,            // _In_ int  nIndex,
    (LONG)PluginWinampProc  // _In_ LONG dwNewLong
  );

  return 0;
}

static int UnSubclassWinamp(HWND hWnd)
{
  if (NULL==hWnd||NULL==WinampProc) {
    DMESSAGE("failing to unsubclass winamp window");
    return -1;
  }

  SetWindowLong(
    hWnd,                   // _In_ HWND hWnd,
    GWL_WNDPROC,            // _In_ int  nIndex,
    (LONG)WinampProc        // _In_ LONG dwNewLong
  );

  return 0;
}
#endif // ]
#endif // ]

#if defined (YASAPI_API) // [
///////////////////////////////////////////////////////////////////////////////
static int YasapiNtApiOpen(int samplerate, int numchannels, int bitspersamp,
    int *pnMaxLatency, int *pnFramesDevice)
{
  return PlayerOpen(&player,samplerate,numchannels,bitspersamp,pnMaxLatency,
      pnFramesDevice);
}

static int YasapiNtApiWrite(int msTimeout, const char *buf, int len)
{
  return EngineApiWrite(&player.engine,msTimeout,buf,len);
}

static int YasapiNtApiEndOfTrack(int msTimeout)
{
  return EngineApiEndOfTrack(&player.engine,msTimeout);
}

static void YasapiNtApiKill(void)
{
  EngineApiKill(&player.engine);
}

#define YASPI_EXPORT_API
#if defined (YASPI_EXPORT_API) // [
YASAPI_EXPORT
#endif // ]
const YasapiNtApi *GetYasapiNtApi(void)
{
  static const YasapiNtApi api={
    .nVersion=eYasapiNtApiVersion,
    .Open=YasapiNtApiOpen,
    .Write=YasapiNtApiWrite,
    .EndOfTrack=YasapiNtApiEndOfTrack,
    .Kill=YasapiNtApiKill,
  };

  return &api;
}
#endif // ]
