/*
 * yasapi_engine_common.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <stdint.h>

///////////////////////////////////////////////////////////////////////////////
#if ! defined (AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM) // [
#define AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM 0x80000000
#endif // ]

#if ! defined (AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY) // [
#define AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY 0x08000000
#endif // ]

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_OPTION_GAPLESS) // [
static int EngineCreateGaplessSession(Engine *pEngine);
static int EngineCreateParams(Engine *pEngine);
#endif // ]
static int EngineCreateBase(Engine *pEngine, Player *pPlayer);
#if defined (YASAPI_OPTION_GAPLESS) // [
static int EngineCreateOptions(Engine *pEngine, EngineOptions *pOptions);
#else // ] [
static int EngineCreateOptions(Engine *pEngine);
#endif // ]
static int EngineGetBufferDuration(Engine *pEngine);
#if defined (YASAPI_RING_REALLOC) // [
static int EngineCreateRing(Engine *pEngine, DWORD dwChunkSizeSource,
    ERingAlloc eRingAlloc);
#else // ] [
static int EngineCreateRing(Engine *pEngine, DWORD dwChunkSizeSource);
#endif // ]
#if ! defined (YASAPI_TIME) // [
static int EngineNotPlaying(Engine *pEngine);
static double EngineOutputTimeInt(Engine *pEngine);
#endif // ]

#if defined (YASAPI_DEBUG) // [
///////////////////////////////////////////////////////////////////////////////
static const char *BitdepthMnemonic(EBitdepth eBitdepth)
{
  switch (eBitdepth) {
  case eBitdepthDisabled:
    return "eBitdepthDisabled";
  case eBitdepth16:
    return "eBitdepth16";
  case eBitdepth24:
    return "eBitdepth24";
  case eBitdepth32:
    return "eBitdepth32";
  default:
    return "undefined";
  }
}
#endif // ]

///////////////////////////////////////////////////////////////////////////////
#if defined (YASAPI_OPTION_GAPLESS) // [
int EngineCreate(Engine *pEngine, Player *pPlayer, int samplerate,
    int numchannels, int bitspersamp, int *pnMaxLatency, int *pnFramesDevice)
{
  int code=-1;
  Gtk *pGtk=&pEngine->pPlayer->gtk;
  int bIncGaplessBroken=pEngine->options.common.bGapless;
  const EEngine eState=pEngine->eState;
  int bStart=0;
  EngineOptions options;
  //int bSurround=pEngine->options.common.bSurround;
  int bSurround;
  WAVEFORMATEXTENSIBLE source;
  WAVEFORMATEXTENSIBLE target;
  WORD nChannelsTarget;
  WORD wBitsPerSampleTarget;
  int bOptionsChanged;
  int bTargetChanged;

DVWRITELN("### %s ####################################",__func__);
  /////////////////////////////////////////////////////////////////////////////
  memset(&options,0,sizeof options);
  memset(&source,0,sizeof source);
  memset(&target,0,sizeof target);

  /////////////////////////////////////////////////////////////////////////////
  if (eState<eEngineBase&&EngineCreateBase(pEngine,pPlayer)<0) {
    DMESSAGE("creating engine base");
    EngineKill(pEngine,eState);
    goto enoinit;
  }

  EnterCriticalSection(&pEngine->cs); // {

  /////////////////////////////////////////////////////////////////////////////
  if (EngineCreateOptions(pEngine,&options)<0) {
    DMESSAGE("creating options");
    EngineKill(pEngine,eState);
    goto exit;
  }

  bOptionsChanged=memcmp(&pEngine->options,&options,sizeof pEngine->options);

  /////////////////////////////////////////////////////////////////////////////
  bSurround=options.common.bSurround;

  if (WfxCreate(&source,samplerate,numchannels,bitspersamp,bSurround)<0) {
    DVMESSAGE("unknown format %d/%d/%d",samplerate,numchannels,
        bitspersamp);
    EngineKill(pEngine,eState);
    goto exit;
  }

  if (1==numchannels&&pEngine->options.common.eMono2Stereo)
    nChannelsTarget=2;
  else
    nChannelsTarget=numchannels;

  switch (options.device.eBitdepth) {
  case eBitdepth16:
    wBitsPerSampleTarget=16u;
    break;
  case eBitdepth24:
    wBitsPerSampleTarget=24u;
    break;
  case eBitdepth32:
    wBitsPerSampleTarget=32u;
    break;
  default:
    wBitsPerSampleTarget=bitspersamp;
    break;
  }

  DVWRITELN("wBitsPerSampleTarget: %u",wBitsPerSampleTarget);

  if (WfxCreate(&target,samplerate,nChannelsTarget,wBitsPerSampleTarget,
      bSurround)<0) {
    DVMESSAGE("unknown format %d/%u/%u",samplerate,nChannelsTarget,
        wBitsPerSampleTarget);
    EngineKill(pEngine,eState);
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  bTargetChanged=memcmp(&pEngine->wfx.target.Format,&target.Format,
      sizeof pEngine->wfx.target.Format);

#if defined (YASAPI_DEBUG) // [
  if (bTargetChanged) {
    DVWRITELN("*** %u/%u/%u %u/%u/%u ***",
        pEngine->wfx.target.Format.nSamplesPerSec,
        pEngine->wfx.target.Format.nChannels,
        pEngine->wfx.target.Format.wBitsPerSample,
        target.Format.nSamplesPerSec,
        target.Format.nChannels,
        target.Format.wBitsPerSample);
  }
#endif // ]

  memcpy(&pEngine->wfx.source,&source,sizeof pEngine->wfx.source);

  /////////////////////////////////////////////////////////////////////////////
DVWRITELN("=== bStart: %d",bStart);
  bStart=bStart||bOptionsChanged;
DVWRITELN("=== bStart: %d (bOptionsChanged)",bStart);
  bStart=bStart||bTargetChanged;
DVWRITELN("=== bStart: %d (bTargetChanged)",bStart);
  bStart=bStart||!options.common.bGapless;
DVWRITELN("=== bStart: %d (bGapless)",bStart);
  bStart=bStart||eEngineBase==pEngine->eState;
DVWRITELN("=== bStart: %d (eEngineBase==pEngine->eState)",bStart);
  bStart=bStart||eEngineEot==pEngine->eState;
DVWRITELN("=== bStart: %d (eEngineEot==pEngine->eState)",bStart);

  if (bStart) {
DWRITELN("+++ bStart");
    ///////////////////////////////////////////////////////////////////////////
    bIncGaplessBroken=bIncGaplessBroken&&!bTargetChanged;
    bIncGaplessBroken=bIncGaplessBroken&&!bOptionsChanged;
    //bIncGaplessBroken=bIncGaplessBroken&&eEngineBase<pEngine->eState;
    bIncGaplessBroken=bIncGaplessBroken&&pEngine->eState<eEngineClose;

    if (bIncGaplessBroken) {
      ++pEngine->pPlayer->stats.uNumGaplessBroken;
      GtkUpdateWidget(pGtk,
          &pGtk->config.config.monitor.player.num_gapless_broken);
    }

    ///////////////////////////////////////////////////////////////////////////
    // We have to wait with the old options set ...
    ResetEvent(pEngine->thread.hEot);

    while (eEngineClose<pEngine->eState) {
      LeaveCriticalSection(&pEngine->cs); // {
      WaitForSingleObject(pEngine->thread.hEot,INFINITE);
      ResetEvent(pEngine->thread.hEot);
      EnterCriticalSection(&pEngine->cs); // }
    };

DVWRITELN("=== %s",EngineState(pEngine));
    //DASSERT_ENGINE_STATE(pEngine,eEngineBase,exit);

    // Now we can copy over the options if needed ...
    if (bOptionsChanged)
      memcpy(&pEngine->options,&options,sizeof pEngine->options);

    if (bTargetChanged)
      memcpy(&pEngine->wfx.target,&target,sizeof pEngine->wfx.target);

    ///////////////////////////////////////////////////////////////////////////
    if (EngineCreateParams(pEngine)<0) {
      DMESSAGE("creating parameter");
      goto exit;
    }

    ///////////////////////////////////////////////////////////////////////////
    ZeroMemory(&pEngine->stats,sizeof pEngine->stats);
    pEngine->stats.uNumTracks=1u;
    ++pEngine->pPlayer->stats.uNumTracks;
    ++pEngine->pPlayer->stats.uNumGapless;

    ///////////////////////////////////////////////////////////////////////////
    if (EngineCreateGaplessSession(pEngine)<0) {
      DMESSAGE("creating gapless sequence");
      goto exit;
    }

DVWRITELN("=== %s",EngineState(pEngine));
    DASSERT_ENGINE_STATE(pEngine,eEngineAudioClientInititial,exit);

#if defined (YASAPI_RING_REALLOC) // [
    GtkUpdateDialog(&pPlayer->gtk,&pPlayer->gtk.config);
#endif // ]
  }
  else {
DWRITELN("=================================================================");
DWRITELN("--- bStart");
#if ! defined (YASAPI_RING_REALLOC) // [
    // We're still in state "eEngineDraining" ...
    EngineKill(pEngine,eEngineTimerStarted);
    ++pEngine->stats.uNumTracks;
    ++pEngine->pPlayer->stats.uNumTracks;
#endif // ]

#if defined (YASAPI_TIME) // [
    pEngine->param.time.lpVtbl->Inc(pEngine);
#else // ] [
    if (!pEngine->options.common.bLatency)
      pEngine->param.msHistory+=EngineOutputTimeInt(pEngine);
#endif // ]

#if ! defined (YASAPI_RING_REALLOC) // [
    if (RingResetWrite(&pEngine->ring,0)<0) {
      DMESSAGE("resetting ring write");
      goto exit;
    }
#endif // ]
  }

  /////////////////////////////////////////////////////////////////////////////
  pEngine->param.bActive=1;
#if ! defined (YASAPI_RING_REALLOC) // [
  GtkUpdateDialog(&pPlayer->gtk,&pPlayer->gtk.config);
#endif // ]

  if (pnMaxLatency)
    *pnMaxLatency=pEngine->options.device.xMaxLatencyAssumed+0.5;

  if (pnFramesDevice)
    *pnFramesDevice=pEngine->param.uNumFramesDevice;

  code=0;
exit:
  LeaveCriticalSection(&pEngine->cs); // }
enoinit:
  return code;
}

static int EngineCreateGaplessSession(Engine *pEngine)
#else // ] [
int EngineCreate(Engine *pEngine, Player *pPlayer, int samplerate,
    int numchannels, int bitspersamp)
#endif // ]
{
#if defined (YASAPI_OPTION_GAPLESS) // [
  int code=-1;
#else // ] [
  int nMaxLatency=0;
#endif // ]
  const EEngine eState=pEngine->eState;

#if defined (YASAPI_OPTION_GAPLESS) // [
  /////////////////////////////////////////////////////////////////////////////
  //DASSERT_ENGINE_STATE(pEngine,eEngineBase,exit);

  while (pEngine->eState<eEngineClose)
    ++pEngine->eState;
#if 0 // [
  /////////////////////////////////////////////////////////////////////////////
//#if defined (__GNUC__) // [
//#pragma GCC diagnostic ignored "-Wmaybe-uninitialized"
//#pragma GCC diagnostic push
//#endif // ]
//  pEngine->wfx.source=*pSource;
//  pEngine->wfx.target=*pTarget;
//#if defined (__GNUC__) // [
//#pragma GCC diagnostic pop
//#endif // ]
#endif // ]
#else // ] [
  int bSurround=pEngine->options.common.bSurround;
  WORD nChannelsTarget;
  WORD wBitsPerSampleTarget;
  Wasapi *pWasapi;

  /////////////////////////////////////////////////////////////////////////////
  if (eState<eEngineBase&&EngineCreateBase(pEngine,pPlayer)<0) {
    DMESSAGE("creating engine base");
    EngineKill(pEngine,eState);
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (!(pWasapi=PlayerWasapi(pPlayer))) {
    DMESSAGE("accessing WASAPI");
    EngineDestroy(pEngine);
    goto exit;
  }

  pEngine->pDevice=pWasapi->selected.pDevice->pDevice;

  /////////////////////////////////////////////////////////////////////////////
  ZeroMemory(&pEngine->param,sizeof pEngine->param);

#if defined (YASAPI_TIME) // [
  /////////////////////////////////////////////////////////////////////////////
  if (pEngine->options.common.bLatency)
    pEngine->param.time.lpVtbl=&gcIEngineTimeNullVtbl;
  else
    pEngine->param.time.lpVtbl=&gcIEngineTimeClockVtbl;
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  ZeroMemory(&pEngine->stats,sizeof pEngine->stats);
  pEngine->stats.uNumTracks=1u;
  ++pEngine->pPlayer->stats.uNumTracks;
#if defined (YASAPI_TIME) // [
  pEngine->param.time.lpVtbl->Set(pEngine,0);
#else // ] [
  pEngine->param.time.msHistory=0.0;
#endif // ]

  if (EngineCreateOptions(pEngine)<0) {
    DMESSAGE("creating options");
    EngineKill(pEngine,eState);
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (WfxCreate(&pEngine->wfx.source,samplerate,numchannels,bitspersamp,
      bSurround)<0) {
    DVMESSAGE("unknown format %d/%d/%d",samplerate,numchannels,
        bitspersamp);
    EngineKill(pEngine,eState);
    goto exit;
  }

  if (1==numchannels&&pEngine->options.common.eMono2Stereo)
    nChannelsTarget=2;
  else
    nChannelsTarget=numchannels;

  switch (pEngine->options.device.eBitdepth) {
  case eBitdepth16:
    wBitsPerSampleTarget=16u;
    break;
  case eBitdepth24:
    wBitsPerSampleTarget=24u;
    break;
  case eBitdepth32:
    wBitsPerSampleTarget=32u;
    break;
  default:
    wBitsPerSampleTarget=bitspersamp;
    break;
  }

  if (WfxCreate(&pEngine->wfx.target,samplerate,nChannelsTarget,
      wBitsPerSampleTarget,bSurround)<0) {
    DVMESSAGE("unknown format %d/%u/%u",samplerate,nChannelsTarget,
        wBitsPerSampleTarget);
    EngineKill(pEngine,eState);
    goto exit;
  }
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  if (EngineCreateAudioClient(pEngine,eEngineAudioClientInititial)<0) {
    DMESSAGE("creating audio client");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DWRITELN(EngineState(pEngine));
  DASSERT_ENGINE_STATE(pEngine,eEngineAudioClientInititial,exit);

#if defined (YASAPI_OPTION_GAPLESS) // [
  code=0;
exit:
  return code;
#else // ] [
  /////////////////////////////////////////////////////////////////////////////
  GtkUpdateDialog(&pPlayer->gtk,&pPlayer->gtk.config);
  nMaxLatency=pEngine->options.device.xMaxLatencyAssumed+0.5;
exit:
  return nMaxLatency;
#endif // ]
}

static int EngineCreateBase(Engine *pEngine, Player *pPlayer)
{
  const EEngine eState=pEngine->eState;
#if defined (YASAPI_ENGINE_ENUMERATOR) // [
  HRESULT hr;
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  pEngine->pPlayer=pPlayer;

  /////////////////////////////////////////////////////////////////////////////
  srand(time(NULL));

  /////////////////////////////////////////////////////////////////////////////
  InitializeCriticalSection(&pEngine->cs);
  ++pEngine->eState;
  DASSERT(eEngineCriticalSection==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  if (!(pEngine->timer.hEvent=CreateEventW(NULL,FALSE,FALSE,NULL))) {
    DMESSAGE("creating timer event");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineTimerEvent==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  if (!(pEngine->timer.hTimer=CreateWaitableTimerW(NULL,FALSE,NULL))) {
    DMESSAGE("creating waitable timer");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineTimer==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  if (!(pEngine->thread.hProducer=CreateEventW(NULL,FALSE,FALSE,NULL))) {
    DMESSAGE("creating producer event");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineEventProducer==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  if (!(pEngine->thread.hConsumer=CreateEventW(NULL,FALSE,FALSE,NULL))) {
    DMESSAGE("creating consumer event");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineEventConsumer==pEngine->eState,exit);

#if defined (YASAPI_OPTION_GAPLESS) // [
  /////////////////////////////////////////////////////////////////////////////
  // This is a MANUAL RESET event!
  if (!(pEngine->thread.hEot=CreateEventW(NULL,TRUE,FALSE,NULL))) {
    DMESSAGE("creating eot event");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineEventEot==pEngine->eState,exit);
#endif // ]

#if defined (YASAPI_API) // [
  /////////////////////////////////////////////////////////////////////////////
  // This is a MANUAL RESET event!
  if (!(pEngine->thread.hApi=CreateEventW(NULL,TRUE,FALSE,NULL))) {
    DMESSAGE("creating api event");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineEventApi==pEngine->eState,exit);
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  EnterCriticalSection(&pEngine->cs); // {
  pEngine->thread.pCommand=NULL;

  pEngine->thread.hThread=CreateThread(
    NULL,     // _In_opt_  LPSECURITY_ATTRIBUTES  lpThreadAttributes,
    0ul,      // _In_      SIZE_T                 dwStackSize,
    EngineThreadProc,
              // _In_      LPTHREAD_START_ROUTINE lpStartAddress,
    pEngine,  // _In_opt_  LPVOID                 lpParameter,
    0ul,      // _In_      DWORD                  dwCreationFlags,
    NULL      // _Out_opt_ LPDWORD                lpThreadId
  );

  if (!pEngine->thread.hThread) {
    DMESSAGE("creating thread");
    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineThread==pEngine->eState,exit);

  while (pEngine->eState<eEngineThreadRunning) {
    LeaveCriticalSection(&pEngine->cs); // {
    WaitForSingleObject(pEngine->thread.hConsumer,INFINITE);
    EnterCriticalSection(&pEngine->cs); // }
  }

  pEngine->thread.pCommand=NULL;
  LeaveCriticalSection(&pEngine->cs); // }

#if defined (YASAPI_ENGINE_ENUMERATOR) // [
  /////////////////////////////////////////////////////////////////////////////
  hr=CoCreateInstance(
    &CLSID_MMDeviceEnumerator,    // _In_  REFCLSID  rclsid,
    NULL,                         // _In_  LPUNKNOWN pUnkOuter,
    CLSCTX_ALL,                   // _In_  DWORD     dwClsContext,
    &IID_IMMDeviceEnumerator,     // _In_  REFIID    riid,
    (void **)&pEngine->pDeviceEnumerator
                                  // _Out_ LPVOID    *ppv
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("CoCreateInstance",REGDB_E_CLASSNOTREG);
    DCMESSAGE("CoCreateInstance",CLASS_E_NOAGGREGATION);
    DCMESSAGE("CoCreateInstance",E_NOINTERFACE);
    DCMESSAGE("CoCreateInstance",E_POINTER);
    DDMESSAGE("CoCreateInstance",hr);
    }

    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineDeviceEnumerator==pEngine->eState,exit);

  /////////////////////////////////////////////////////////////////////////////
  hr=pEngine->pDeviceEnumerator->lpVtbl->GetDefaultAudioEndpoint(
    pEngine->pDeviceEnumerator,
    eRender,                      // [in]  EDataFlow dataFlow,
    eConsole,                     // [in]  ERole     role,
    &pEngine->pDevice             // [out] IMMDevice **ppDevice
  );

#if defined (_MSC_VER) // [
  if (E_NOTFOUND==hr) {
    DMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint failed with"
        " E_NOTFOUND");
    EngineKill(pEngine,eState);
    goto exit;
  }
  else
#endif // ]
  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint",E_POINTER);
    DCMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint",E_INVALIDARG);
#if ! defined (_MSC_VER) // [
    DCMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint",E_NOTFOUND);
#endif // ]
    DCMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint",E_OUTOFMEMORY);
    DDMESSAGE("IMMDeviceEnumerator::GetDefaultAudioEndpoint",hr);
    }

    EngineKill(pEngine,eState);
    goto exit;
  }

  ++pEngine->eState;
  DASSERT(eEngineDevice==pEngine->eState,exit);
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
exit:
  DWRITELN(EngineState(pEngine));
  return eState<pEngine->eState?0:-1;
}

#if defined (YASAPI_OPTION_GAPLESS) // [
static int EngineCreateOptions(Engine *pEngine, EngineOptions *pOptions)
#else // ] [
static int EngineCreateOptions(Engine *pEngine)
#endif // ]
{
  int code=-1;
#if defined (YASAPI_OPTION_GAPLESS) // [
  OptionsCommon *pCommon=&pOptions->common;
  OptionsDevice *pDevice=&pOptions->device;
#else // ] [
  OptionsCommon *pCommon=&pEngine->options.common;
  OptionsDevice *pDevice=&pEngine->options.device;
#endif // ]
  Wasapi *pWasapi;

  /////////////////////////////////////////////////////////////////////////////
  if (!(pWasapi=PlayerWasapi(pEngine->pPlayer))) {
    DMESSAGE("accessing WASAPI");
    EngineDestroy(pEngine);
    goto exit;
  }

  DVWRITELN("=== %d =================================",
      pWasapi->selected.pDevice-pWasapi->pDevices);
  *pCommon=pWasapi->options.common.current;
  *pDevice=pWasapi->selected.pDevice->pMember->current;
  DVWRITELN("=== %s",BitdepthMnemonic(pDevice->eBitdepth));

#if ! defined (YASAPI_OPTION_GAPLESS) // [
  /////////////////////////////////////////////////////////////////////////////
  pEngine->param.pFriendlyName
      =pWasapi->selected.pDevice->pFriendlyNameOriginal;

  if (pDevice->eStrategy) {
    pEngine->param.dwStreamFlags=AUDCLNT_STREAMFLAGS_EVENTCALLBACK;
    DWRITELN("AUDCLNT_STREAMFLAGS_EVENTCALLBACK");
  }
  else
    pEngine->param.dwStreamFlags=0ul;

  if (pDevice->eMode) {
    pEngine->param.eShareMode=AUDCLNT_SHAREMODE_EXCLUSIVE;
    DWRITELN("AUDCLNT_SHAREMODE_EXCLUSIVE");
  }
  else {
    pEngine->param.eShareMode=AUDCLNT_SHAREMODE_SHARED;
    DWRITELN("AUDCLNT_SHAREMODE_SHARED");

    if (!pDevice->bAutoConvertPCM)
      goto shared;

    pEngine->param.dwStreamFlags|=AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM;
    DWRITELN("AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM");

    if (! pDevice->bSRCDefaultQuality)
      goto shared;

    pEngine->param.dwStreamFlags|=AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY;
    DWRITELN("AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY");
  shared: ;
  }

  DVWRITELN("dwStreamFlags=%0lx",pEngine->param.dwStreamFlags);
  pEngine->param.bDoubleBuffered=pDevice->eMode&&pDevice->eStrategy;

  /////////////////////////////////////////////////////////////////////////////
  switch (pEngine->options.common.eVolume) {
  case eVolumeSoftware:
    pEngine->param.pIEngineVolumeVtbl=&gcIEngineVolumeSoftwareVtbl;
    break;
  case eVolumeWasapi:
    switch (pEngine->options.device.eMode) {
    case eModeShared:
      pEngine->param.pIEngineVolumeVtbl=&gcIEngineVolumeSimpleAudioVtbl;
      break;
    case eModeExclusive:
      pEngine->param.pIEngineVolumeVtbl=&gcIEngineVolumeAudioEndpointVtbl;
      break;
    default:
      pEngine->param.pIEngineVolumeVtbl=&gcIEngineVolumeSoftwareVtbl;
      break;
    }

    break;
  default:
    pEngine->param.pIEngineVolumeVtbl=&gcIEngineVolumeNullVtbl;
    break;
  };
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}

#if defined (YASAPI_OPTION_GAPLESS) // [
static int EngineCreateParams(Engine *pEngine)
{
  int code=-1;
#if 0 // [
  OptionsCommon *pCommon=&pEngine->options.common;
#endif // ]
  OptionsDevice *pDevice=&pEngine->options.device;
  Wasapi *pWasapi;

  /////////////////////////////////////////////////////////////////////////////
  if (!(pWasapi=PlayerWasapi(pEngine->pPlayer))) {
    DMESSAGE("accessing WASAPI");
    EngineDestroy(pEngine);
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  pEngine->pDevice=pWasapi->selected.pDevice->pDevice;

  /////////////////////////////////////////////////////////////////////////////
  ZeroMemory(&pEngine->param,sizeof pEngine->param);
#if defined (YASAPI_TIME) // [
  if (pEngine->options.common.bLatency)
    pEngine->param.time.lpVtbl=&gcIEngineTimeNullVtbl;
  else
    pEngine->param.time.lpVtbl=&gcIEngineTimeClockVtbl;

  pEngine->param.time.lpVtbl->Set(pEngine,0);
#else // ] [
  pEngine->param.msHistory=0.0;
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  pEngine->param.pFriendlyName
      =pWasapi->selected.pDevice->pFriendlyNameOriginal;

  if (pDevice->eStrategy) {
    pEngine->param.dwStreamFlags=AUDCLNT_STREAMFLAGS_EVENTCALLBACK;
    DWRITELN("AUDCLNT_STREAMFLAGS_EVENTCALLBACK");
  }
  else
    pEngine->param.dwStreamFlags=0ul;

  if (pDevice->eMode) {
    pEngine->param.eShareMode=AUDCLNT_SHAREMODE_EXCLUSIVE;
    DWRITELN("AUDCLNT_SHAREMODE_EXCLUSIVE");
  }
  else {
    pEngine->param.eShareMode=AUDCLNT_SHAREMODE_SHARED;
    DWRITELN("AUDCLNT_SHAREMODE_SHARED");

    if (!pDevice->bAutoConvertPCM)
      goto shared;

    pEngine->param.dwStreamFlags|=AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM;
    DWRITELN("AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM");

    if (! pDevice->bSRCDefaultQuality)
      goto shared;

    pEngine->param.dwStreamFlags|=AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY;
    DWRITELN("AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY");
  shared: ;
  }

  DVWRITELN("dwStreamFlags=%0lx",pEngine->param.dwStreamFlags);
  pEngine->param.bDoubleBuffered=pDevice->eMode&&pDevice->eStrategy;

  /////////////////////////////////////////////////////////////////////////////
  switch (pEngine->options.common.eVolume) {
  case eVolumeSoftware:
    pEngine->param.pIEngineVolumeVtbl=&gcIEngineVolumeSoftwareVtbl;
    break;
  case eVolumeWasapi:
    switch (pEngine->options.device.eMode) {
    case eModeShared:
      pEngine->param.pIEngineVolumeVtbl=&gcIEngineVolumeSimpleAudioVtbl;
      break;
    case eModeExclusive:
      pEngine->param.pIEngineVolumeVtbl=&gcIEngineVolumeAudioEndpointVtbl;
      break;
    default:
      pEngine->param.pIEngineVolumeVtbl=&gcIEngineVolumeSoftwareVtbl;
      break;
    }

    break;
  default:
    pEngine->param.pIEngineVolumeVtbl=&gcIEngineVolumeNullVtbl;
    break;
  };

  code=0;
exit:
  return code;
}
#endif // ]

static int EngineGetBufferDuration(Engine *pEngine)
{
  int code=-1;
  REFERENCE_TIME hnsBufferDuration;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
  if (pEngine->options.device.eMode) {
    hr=pEngine->pAudioClient->lpVtbl->GetDevicePeriod(pEngine->pAudioClient,
      NULL,       // [out] REFERENCE_TIME *phnsDefaultDevicePeriod,
      &hnsBufferDuration
                  // [out] REFERENCE_TIME *phnsMinimumDevicePeriod
    );
  }
  else {
    hr=pEngine->pAudioClient->lpVtbl->GetDevicePeriod(pEngine->pAudioClient,
      &hnsBufferDuration,
                  // [out] REFERENCE_TIME *phnsDefaultDevicePeriod,
      NULL        // [out] REFERENCE_TIME *phnsMinimumDevicePeriod
    );
  }

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClient::GetDevicePeriod",AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClient::GetDevicePeriod",AUDCLNT_E_SERVICE_NOT_RUNNING);
    DCMESSAGE("IAudioClient::GetDevicePeriod",E_POINTER);
    DDMESSAGE("IAudioClient::GetDevicePeriod",hr);
    }

    goto exit;
  }

  DVWRITELN("*** hnsBufferDuration=%I64d ***",hnsBufferDuration);
  pEngine->param.hnsBufferDuration=hnsBufferDuration;

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}

int EngineCreateAudioClient(Engine *pEngine, const EEngine eState)
{
  int code=-1;
  int nAlignment=0;
  UINT32 uNumFrames;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
activate:
  hr=pEngine->pDevice->lpVtbl->Activate(pEngine->pDevice,
    &IID_IAudioClient,      // [in]  REFIID      iid,
    CLSCTX_ALL,             // [in]  DWORD       dwClsCtx,
    NULL,                   // [in]  PROPVARIANT *pActivationParams,
    (void **)&pEngine->pAudioClient
                            // [out] void        **ppInterface
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IMMDevice::Activate",E_NOINTERFACE);
    DCMESSAGE("IMMDevice::Activate",E_POINTER);
    DCMESSAGE("IMMDevice::Activate",E_INVALIDARG);
    DCMESSAGE("IMMDevice::Activate",E_OUTOFMEMORY);
    DCMESSAGE("IMMDevice::Activate",AUDCLNT_E_DEVICE_INVALIDATED);
    DDMESSAGE("IMMDevice::Activate",hr);
    }

    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (0ll==pEngine->param.hnsBufferDuration) {
    if (eEngineAudioClientInititial!=eState) {
      DMESSAGE("internal");
      goto exit;
    }
    else if (EngineGetBufferDuration(pEngine)<0) {
      DMESSAGE("retrieving buffer duration");
      goto exit;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  if (pEngine->options.device.eStrategy) {
    hr=pEngine->pAudioClient->lpVtbl->Initialize(pEngine->pAudioClient,
      pEngine->param.eShareMode,
                            // [in]       AUDCLNT_SHAREMODE ShareMode,
      pEngine->param.dwStreamFlags,
                            // [in]       DWORD             StreamFlags,
      pEngine->param.hnsBufferDuration,
                            // [in]       REFERENCE_TIME    hnsBufferDuration,
      pEngine->param.hnsBufferDuration,
                            // [in]       REFERENCE_TIME    hnsPeriodicity,
      &pEngine->wfx.target.Format,
                            // [in] const WAVEFORMATEX      *pFormat,
      NULL                  // [in]       LPCGUID           AudioSessionGuid
    );
  }
  else {
    hr=pEngine->pAudioClient->lpVtbl->Initialize(pEngine->pAudioClient,
      pEngine->param.eShareMode,
                            // [in]       AUDCLNT_SHAREMODE ShareMode,
      pEngine->param.dwStreamFlags,
                            // [in]       DWORD             StreamFlags,
      pEngine->param.hnsBufferDuration,
                            // [in]       REFERENCE_TIME    hnsBufferDuration,
      0ull,                 // [in]       REFERENCE_TIME    hnsPeriodicity,
      &pEngine->wfx.target.Format,
                            // [in] const WAVEFORMATEX      *pFormat,
      NULL                  // [in]       LPCGUID           AudioSessionGuid
    );
  }

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClient::Initialize",CO_E_NOTINITIALIZED);
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_ALREADY_INITIALIZED);
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_WRONG_ENDPOINT_TYPE);
    case AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED:
      if (pEngine->options.device.bAlignmentDance&&!nAlignment) {
        // https://blogs.msdn.microsoft.com/matthew_van_eerde/2009/04/03/sample-wasapi-exclusive-mode-event-driven-playback-app-including-the-hd-audio-alignment-dance/
        hr=pEngine->pAudioClient->lpVtbl->GetBufferSize(pEngine->pAudioClient,
          &uNumFrames                             // UINT32 *pNumBufferFrames
        );

        pEngine->param.hnsBufferDuration=
            10.0*1000                             // (hns / ms) *
            *1000                                 // (ms / s) *
            *uNumFrames                           // frames /
            /pEngine->wfx.target.Format.nSamplesPerSec
                                                  // (frames / s)
            +0.5                                  // rounding
            ;

        DVWRITELN("*** uNumFrames (first pass)=%u ***",uNumFrames);
        DVWRITELN("*** hnsBufferDuration (first pass)=%I64d ***",
            pEngine->param.hnsBufferDuration);

        --pEngine->eState;
        DVWRITELN("--< %s",EngineState(pEngine));
        hr=pEngine->pAudioClient->lpVtbl->Release(pEngine->pAudioClient);
        pEngine->pAudioClient=NULL;
        ++nAlignment;
        goto activate;
      }

      DMESSAGE("IAudioClient::Initialize failed with"
          " AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED");
      break;
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_BUFFER_SIZE_ERROR);
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_CPUUSAGE_EXCEEDED);
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_DEVICE_IN_USE);
    //DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_ENDPOINT_CREATE_FAILED);
    case AUDCLNT_E_ENDPOINT_CREATE_FAILED:
      if (1==pEngine->wfx.target.Format.nChannels
          &&!pEngine->options.common.eMono2Stereo) {
        DMESSAGE("IAudioClient::Initialize failed with"
            " AUDCLNT_E_ENDPOINT_CREATE_FAILED"
            " (consider switching on mono to stereo convertion)");
        goto exit;
      }

#if 0 // [
      DVMESSAGE("IAudioClient::Initialize failed with"
          " AUDCLNT_E_ENDPOINT_CREATE_FAILED (nChannels: %u)",
          pEngine->wfx.target.Format.nChannels);
#else // ] [
      DVMESSAGE("IAudioClient::Initialize failed with"
          " AUDCLNT_E_ENDPOINT_CREATE_FAILED (nChannels: %u"
          ", nSamplesPerSec: %lu, nBlockAlign: %u, wBitsPerSample: %u)",
          pEngine->wfx.target.Format.nChannels,
          pEngine->wfx.target.Format.nSamplesPerSec,
          pEngine->wfx.target.Format.nBlockAlign,
          pEngine->wfx.target.Format.wBitsPerSample);
#endif // ]
      break;
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_INVALID_DEVICE_PERIOD);
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_UNSUPPORTED_FORMAT);
    DCMESSAGE("IAudioClient::Initialize",
        AUDCLNT_E_EXCLUSIVE_MODE_NOT_ALLOWED);
    DCMESSAGE("IAudioClient::Initialize",
        AUDCLNT_E_BUFDURATION_PERIOD_NOT_EQUAL);
    DCMESSAGE("IAudioClient::Initialize",AUDCLNT_E_SERVICE_NOT_RUNNING);
    DCMESSAGE("IAudioClient::Initialize",E_POINTER);
    DCMESSAGE("IAudioClient::Initialize",E_INVALIDARG);
    DCMESSAGE("IAudioClient::Initialize",E_OUTOFMEMORY);
    DDMESSAGE("IAudioClient::Initialize",hr);
    }

    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (0u==pEngine->param.uNumFramesDevice) {
    if (eEngineAudioClientInititial==eState) {
      hr=pEngine->pAudioClient->lpVtbl->GetBufferSize(pEngine->pAudioClient,
        &uNumFrames       // UINT32 *pNumBufferFrames
      );

      if (FAILED(hr)) {
        switch (hr) {
        DCMESSAGE("IAudioClient::GetBufferSize",AUDCLNT_E_NOT_INITIALIZED);
        DCMESSAGE("IAudioClient::GetBufferSize",AUDCLNT_E_DEVICE_INVALIDATED);
        DCMESSAGE("IAudioClient::GetBufferSize",AUDCLNT_E_SERVICE_NOT_RUNNING);
        DCMESSAGE("IAudioClient::GetBufferSize",E_POINTER);
        DDMESSAGE("IAudioClient::GetBufferSize",hr);
        }

        goto exit;
      }

      pEngine->param.uNumFramesDevice=uNumFrames;

      if (!pEngine->options.device.eStrategy) {
        pEngine->param.DueTime.QuadPart
            =-0.5*YASAPI_FRAMES2HNS(uNumFrames,&pEngine->wfx.target)-0.5;

        if (pEngine->options.device.bTimerPeriodic) {
          pEngine->param.lPeriod
              =0.5*YASAPI_FRAMES2MS(uNumFrames,&pEngine->wfx.target)+0.5;
        }
        else
          pEngine->param.lPeriod=0l;
      }
      else {
        pEngine->param.DueTime.QuadPart=0ll;
        pEngine->param.lPeriod=0l;
      }
    }
  }

//#if ! defined (YASAPI_ENGINE_REGISTER_WHEN_STARING) // [
  /////////////////////////////////////////////////////////////////////////////
  if (pEngine->options.device.eStrategy&&!pEngine->options.common.bDisconnect) {
    hr=pEngine->pAudioClient->lpVtbl->SetEventHandle(pEngine->pAudioClient,
      pEngine->timer.hEvent   // [in] HANDLE eventHandle
    );

    if (FAILED(hr)) {
      switch (hr) {
      DCMESSAGE("IAudioClient::SetEventHandle",E_INVALIDARG);
      DCMESSAGE("IAudioClient::SetEventHandle",
          AUDCLNT_E_EVENTHANDLE_NOT_EXPECTED);
      DCMESSAGE("IAudioClient::SetEventHandle",AUDCLNT_E_NOT_INITIALIZED);
      DCMESSAGE("IAudioClient::SetEventHandle",AUDCLNT_E_DEVICE_INVALIDATED);
      DCMESSAGE("IAudioClient::SetEventHandle",AUDCLNT_E_SERVICE_NOT_RUNNING);
      DDMESSAGE("IAudioClient::SetEventHandle",hr);
      }
  
      EngineKill(pEngine,eState);
      goto exit;
    }

    DWRITELN("event handle registered");
  }
//#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}

///////////////////////////////////////////////////////////////////////////////
void EngineKill(Engine *pEngine, const EEngine eState)
{
  Gtk *pGtk=&pEngine->pPlayer->gtk;

  switch (pEngine->eState) {
  case eEngineEot:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineEot==pEngine->eState,exit);
    --pEngine->eState;
    // Intentional fall-through.
  case eEngineDrainingFinal:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineDrainingFinal==pEngine->eState,exit);
    --pEngine->eState;
    // Intentional fall-through.
  case eEngineDraining:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineDraining==pEngine->eState,exit);
    --pEngine->eState;
    // Intentional fall-through.
  case eEngineTimerStarted:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineTimerStarted==pEngine->eState,exit);
    --pEngine->eState;

    if (!pEngine->options.device.eStrategy)
      CancelWaitableTimer(pEngine->timer.hTimer);

    // Intentional fall-through.
  case eEngineAudioClientStarted:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineAudioClientStarted==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->pAudioClient->lpVtbl->Stop(pEngine->pAudioClient);

    if (pEngine->param.bDoubleBuffered)
      pEngine->pAudioClient->lpVtbl->Reset(pEngine->pAudioClient);
    // Intentional fall-through.
  case eEnginePauseConnected:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEnginePauseConnected==pEngine->eState,exit);
    --pEngine->eState;
    // Intentional fall-through.
  case eEngineAudioRenderClient:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineAudioRenderClient==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->pAudioRenderClient->lpVtbl->Release(pEngine->pAudioRenderClient);
    pEngine->pAudioRenderClient=NULL;
    // Intentional fall-through.
  case eEngineAudioClock:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineAudioClock==pEngine->eState,exit);

#if defined (YASAPI_TIME) // [
    pEngine->param.time.lpVtbl->ClockDestroy(pEngine);
#else // ] [
    if (!pEngine->options.common.bLatency) {
      pEngine->param.msHistory+=EngineOutputTimeInt(pEngine);
      pEngine->pAudioClock->lpVtbl->Release(pEngine->pAudioClock);
      pEngine->pAudioClock=NULL;
    }
#endif // ]

    --pEngine->eState;
    // Intentional fall-through.
  case eEngineVolume:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineVolume==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->param.pIEngineVolumeVtbl->Destroy(pEngine);
    // Intentional fall-through.
  case eEngineAudioClient:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineAudioClient==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->pAudioClient->lpVtbl->Release(pEngine->pAudioClient);
    pEngine->pAudioClient=NULL;
    // Intentional fall-through.
  case eEnginePauseDisconnected:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEnginePauseDisconnected==pEngine->eState,exit);
    --pEngine->eState;
    // Intentional fall-through.
  case eEngineRing:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineRing==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->monitor.xWritten=0.0;
    pEngine->monitor.xPadding=0.0;
    RingDestroy(&pEngine->ring);

    ZeroMemory(&pEngine->param,sizeof pEngine->param);
    ZeroMemory(&pEngine->stats,sizeof pEngine->stats);
    pEngine->stats.uNumTracks=UINT32_MAX;
    pEngine->stats.uNumDropouts=UINT32_MAX;
#if 0 && defined (YASAPI_OPTION_GAPLESS) // [
    pEngine->stats.uNumGaplessBroken=UINT32_MAX;
#endif // ]
#if defined (YASAPI_RING2) // [
    pEngine->param.uFramesPlaybackBegin=0u;
    pEngine->monitor.xStart=0.0;
#endif // ]
    GtkUpdateDialog(pGtk,&pGtk->config);
    // Intentional fall-through.
  case eEngineAudioClientInititial:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineAudioClientInititial==pEngine->eState,exit);
    --pEngine->eState;

    if (pEngine->pAudioClient) {
      pEngine->pAudioClient->lpVtbl->Release(pEngine->pAudioClient);
      pEngine->pAudioClient=NULL;
    }
    // Intentional fall-through.
#if defined (YASAPI_OPTION_GAPLESS) // [
  case eEngineClose:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineClose==pEngine->eState,exit);
    --pEngine->eState;
#if 0 // [
    memset(&pEngine->wfx.source,0,sizeof pEngine->wfx.source);
    memset(&pEngine->wfx.target,0,sizeof pEngine->wfx.target);
    memset(&pEngine->param,0,sizeof pEngine->param);
#endif // ]
    // Intentional fall-through.
#endif // ]
#if defined (YASAPI_ENGINE_ENUMERATOR) // [
  case eEngineDevice:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineDevice==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->pDevice->lpVtbl->Release(pEngine->pDevice);
    pEngine->pDevice=NULL;
    // Intentional fall-through.
  case eEngineDeviceEnumerator:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineDeviceEnumerator==pEngine->eState,exit);
    --pEngine->eState;
    pEngine->pDeviceEnumerator->lpVtbl->Release(pEngine->pDeviceEnumerator);
    pEngine->pDeviceEnumerator=NULL;
    // Intentional fall-through.
#endif // ]
  case eEngineThreadRunning:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineThreadRunning==pEngine->eState,exit);
    --pEngine->eState;
    EngineSend(pEngine,eCommandExit,eEngineNull,1);
    WaitForSingleObject(pEngine->thread.hThread,INFINITE);
#if 0 // [
    memset(&pEngine->wfx.source,0,sizeof pEngine->wfx.source);
    memset(&pEngine->wfx.target,0,sizeof pEngine->wfx.target);
    memset(&pEngine->param,0,sizeof pEngine->param);
#endif // ]
    // Intentional fall-through.
  case eEngineThread:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineThread==pEngine->eState,exit);
    --pEngine->eState;
    CloseHandle(pEngine->thread.hThread);
    pEngine->thread.hThread=NULL;
    // Intentional fall-through.
#if defined (YASAPI_API) // [
  case eEngineEventApi:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineEventApi==pEngine->eState,exit);
    --pEngine->eState;
    CloseHandle(pEngine->thread.hApi);
    pEngine->thread.hApi=NULL;
    // Intentional fall-through.
#endif // ]
#if defined (YASAPI_OPTION_GAPLESS) // [
  case eEngineEventEot:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineEventEot==pEngine->eState,exit);
    --pEngine->eState;
    CloseHandle(pEngine->thread.hEot);
    pEngine->thread.hEot=NULL;
    // Intentional fall-through.
#endif // ]
  case eEngineEventConsumer:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineEventConsumer==pEngine->eState,exit);
    --pEngine->eState;
    CloseHandle(pEngine->thread.hConsumer);
    pEngine->thread.hConsumer=NULL;
    // Intentional fall-through.
  case eEngineEventProducer:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineEventProducer==pEngine->eState,exit);
    --pEngine->eState;
    CloseHandle(pEngine->thread.hProducer);
    pEngine->thread.hProducer=NULL;
    // Intentional fall-through.
  case eEngineTimer:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineTimer==pEngine->eState,exit);
    --pEngine->eState;
    CloseHandle(pEngine->timer.hTimer);
    pEngine->timer.hTimer=NULL;
    // Intentional fall-through.
  case eEngineTimerEvent:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineTimerEvent==pEngine->eState,exit);
    --pEngine->eState;
    CloseHandle(pEngine->timer.hEvent);
    pEngine->timer.hEvent=NULL;
    // Intentional fall-through.
  case eEngineCriticalSection:
    if (eState==pEngine->eState)
      break;

    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineCriticalSection==pEngine->eState,exit);
    --pEngine->eState;
    DeleteCriticalSection(&pEngine->cs);
    // Intentional fall-through.
  default:
    break;
  }

  DWRITELN(EngineState(pEngine));
DEXIT();
  return;
}

void EngineDestroy(Engine *pEngine)
{
  EngineKill(pEngine,eEngineNull);
}

void EngineClose(Engine *pEngine)
{
#if defined (YASAPI_OPTION_GAPLESS) // [
  int bKill=0;
#endif // ]

DVWRITELN("### %s ####################################",__func__);
  if (pEngine->eState<eEngineBase) {
    DMESSAGE("not initialized");
    goto enoinit;
  }

  EnterCriticalSection(&pEngine->cs); // {
  DVWRITELN("=== target: %u/%u/%u ===",
      pEngine->wfx.target.Format.nSamplesPerSec,
      pEngine->wfx.target.Format.nChannels,
      pEngine->wfx.target.Format.wBitsPerSample);
#if defined (YASAPI_OPTION_GAPLESS) // [
DVWRITELN("bKill: %d",bKill);
  bKill=bKill||!pEngine->param.bActive;
DVWRITELN("bKill: %d (!pEngine->param.bActive)",bKill);
  bKill=bKill||!pEngine->options.common.bGapless;
DVWRITELN("bKill: %d (!pEngine->options.common.bGapless)",bKill);
  bKill=bKill||(eEngineBase<pEngine->eState&&pEngine->eState<eEngineDraining&&pEngine->eState<eEngineDraining);
DVWRITELN("bKill: %d (%s)",bKill,EngineState(pEngine));

  if (bKill) {
DWRITELN("=== KILLING ==================================================");
    pEngine->param.bActive=0;
#endif // ]
    EngineSend(pEngine,eCommandKill,eEngineBase,0);
#if defined (YASAPI_OPTION_GAPLESS) // [
#if 1 // [
    memset(&pEngine->wfx.source,0,sizeof pEngine->wfx.source);
    memset(&pEngine->wfx.target,0,sizeof pEngine->wfx.target);
    memset(&pEngine->param,0,sizeof pEngine->param);
#endif // ]
    ++pEngine->eState;
    DASSERT_ENGINE_STATE(pEngine,eEngineClose,exit);
  }
DEXIT();
#endif // ]
  DVWRITELN("=== target: %u/%u/%u ===",
      pEngine->wfx.target.Format.nSamplesPerSec,
      pEngine->wfx.target.Format.nChannels,
      pEngine->wfx.target.Format.wBitsPerSample);
  LeaveCriticalSection(&pEngine->cs); // }
enoinit: ;
}

///////////////////////////////////////////////////////////////////////////////
int EnginePause(Engine *pEngine, int nPause)
{
  int nLastPause=pEngine->param.nLastPause;

  if (nLastPause==nPause)
    goto epause;

  if (pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs); // {

  if (nPause) {
    switch (pEngine->eState) {
    case eEngineStarted:
      if (pEngine->options.common.bDisconnect)
        EngineSend(pEngine,eCommandKill,eEnginePauseDisconnected,0);
      else
        EngineSend(pEngine,eCommandKill,eEnginePauseConnected,0);

      pEngine->param.nLastPause=nPause;
      break;
    default:
      break;
    }
  }
  else {
    switch (pEngine->eState) {
    case eEnginePauseDisconnected:
    case eEnginePauseConnected:
      EngineSend(pEngine,eCommandStart,eEngineNull,0);
      pEngine->param.nLastPause=nPause;
      break;
    default:
      break;
    }
  }

  LeaveCriticalSection(&pEngine->cs); // }
ebase:
epause:
  return nLastPause;
}

int EngineIsPlaying(Engine *pEngine)
{
  int bIsPlaying=0;

  if (pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs); // {

  switch (pEngine->eState) {
  case eEnginePauseDisconnected:
  case eEnginePauseConnected:
    break;
  case eEngineStarted:
    ++pEngine->eState;
    DASSERT(eEngineDraining==pEngine->eState,exit);
    bIsPlaying=1;
    break;
  case eEngineDraining:
  case eEngineDrainingFinal:
#if defined (YASAPI_OPTION_GAPLESS) // [
    bIsPlaying=!pEngine->param.bActive||!pEngine->options.common.bGapless;
#else // ] [
    bIsPlaying=1;
#endif // ]
    break;
  default:
    break;
  }
DEXIT();
#if 0 // [
DVWRITELN("%s (%d)",EngineState(pEngine),bIsPlaying);
#endif // ]
  LeaveCriticalSection(&pEngine->cs); // }
ebase:
  return bIsPlaying;
}

int EngineAvailable(Engine *pEngine)
{
  Ring *pRing=&pEngine->ring;
#if defined (YASAPI_OPTION_GAPLESS) // [
  Gtk *pGtk=&pEngine->pPlayer->gtk;
  ConfigMonitorPlayer *pMonitorPlayer=&pGtk->config.config.monitor.player;
  ConfigWidget *pNumGaplessBroken=&pMonitorPlayer->num_gapless_broken;
#endif // ]
  int nAvailable=0;

  if (pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs); // {

#if 0 && defined (YASAPI_OPTION_GAPLESS) // [
  if (eEngineBase==pEngine->eState) {
    // During "open" we thought this is "gapless". Meanwhile the sesssion was
    // terminated.
    if (EngineCreateGaplessSession(pEngine)<0) {
      DMESSAGE("creating gapless sequence");
      goto exit;
    }

    ++pEngine->pPlayer->stats.uNumGaplessBroken;
    GtkUpdateWidget(pGtk,
        &pGtk->config.config.monitor.player.num_gapless_broken);
  }
#endif // ]

  switch (pEngine->eState) {
#if defined (YASAPI_OPTION_GAPLESS) // [
  case eEngineBase:
    // During "open" we thought this is "gapless". Meanwhile the sesssion was
    // terminated.
    if (EngineCreateGaplessSession(pEngine)<0) {
      DMESSAGE("creating gapless sequence");
      goto exit;
    }

    ++pEngine->pPlayer->stats.uNumGaplessBroken;
    GtkUpdateWidget(pGtk,pNumGaplessBroken);
    nAvailable=INT_MAX;
    break;
#endif // ]
  case eEnginePauseDisconnected:
  case eEnginePauseConnected:
    break;
#if 0 // [
  case eEngineDraining:
  case eEngineDrainingFinal:
    nAvailable=INT_MAX;
    break;
#endif // ]
  default:
    if (pEngine->eState<eEngineRing)
      nAvailable=INT_MAX;
    else
      nAvailable=RingSourceSizeAvailable(pRing);

    break;
  }

#if defined (YASAPI_OPTION_GAPLESS) // [
exit:
#endif // ]
  LeaveCriticalSection(&pEngine->cs); // }
ebase:
//DVWRITELN("nAvailable: %d",nAvailable);
  return nAvailable;
}

#if ! defined (YASAPI_TIME) // [
static int EngineNotPlaying(Engine *pEngine)
{
  int bIsPause=0;

  if (pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs); // {

  switch (pEngine->eState) {
  case eEnginePauseDisconnected:
  case eEnginePauseConnected:
  case eEngineEot:
    bIsPause=1;
    break;
  default:
    break;
  }

  LeaveCriticalSection(&pEngine->cs); // {
ebase:
  return bIsPause;
}

// Return just the raw time without history correction.
static double EngineOutputTimeInt(Engine *pEngine)
{
  double msOutputTime=0.0;
  UINT64 u64Frequency;
  UINT64 u64Position;
  HRESULT hr;

  if (pEngine->eState<eEngineBase||!pEngine->options.common.bLatency) {
    // With "bLatency" we return just 0.
    goto ebase;
  }

  EnterCriticalSection(&pEngine->cs);

  if (EngineNotPlaying(pEngine))
    goto exit;

  if (pEngine->eState<eEngineAudioClock)
    goto exit;

  if (eEnginePauseDisconnected==pEngine->eState)
    goto exit;

  hr=pEngine->pAudioClock->lpVtbl->GetFrequency(pEngine->pAudioClock,
    &u64Frequency     // [out] UINT64 *pu64Frequency
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClock::GetFrequency",E_POINTER);
    DCMESSAGE("IAudioClock::GetFrequency",AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClock::GetFrequency",AUDCLNT_E_SERVICE_NOT_RUNNING);
    DDMESSAGE("IAudioClock::GetFrequency",hr);
    }

    goto exit;
  }

  hr=pEngine->pAudioClock->lpVtbl->GetPosition(pEngine->pAudioClock,
    &u64Position,     // [out] UINT64 *pu64Position
    NULL              // UINT64 *pu64QPCPosition
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioClock::GetPosition",E_POINTER);
    DCMESSAGE("IAudioClock::GetPosition",AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE("IAudioClock::GetPosition",AUDCLNT_E_SERVICE_NOT_RUNNING);
    DDMESSAGE("IAudioClock::GetPosition",hr);
    }

    goto exit;
  }

  msOutputTime+=1000.0*u64Position/u64Frequency;
exit:
  LeaveCriticalSection(&pEngine->cs);
ebase:
  return msOutputTime;
}
#endif // ]

#if defined (YASAPI_TIME) // [
int EngineOutputTime(Engine *pEngine)
#else // ] [
// Return time history corrected.
int EngineOutputTimeExt(Engine *pEngine)
#endif // ]
{
  int ms=0;

  if (pEngine->options.common.bLatency&&pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs); // {

#if defined (YASAPI_TIME) // [
  switch (pEngine->eState) {
  case eEnginePauseDisconnected:
  case eEnginePauseConnected:
  case eEngineEot:
    goto exit;
  default:
    if (pEngine->eState<eEngineAudioClock)
      goto exit;

    break;
  }
#else // ] [
  if (EngineNotPlaying(pEngine))
    goto exit;
#endif // ]

#if defined (YASAPI_TIME) // [
  ms=pEngine->param.time.lpVtbl->Get(pEngine);
#else // ] [
  if (!pEngine->options.common.bLatency)
    ms=EngineOutputTimeInt(pEngine)-pEngine->param.msHistory+0.5;
#endif // ]
exit:
  LeaveCriticalSection(&pEngine->cs); // }
ebase:
  return ms;
}

int EngineWrittenTime(Engine *pEngine)
{
#if defined (YASAPI_RING2) // [
  WAVEFORMATEXTENSIBLE *pTarget=&pEngine->wfx.target;
#else // ] [
  DWORD nAvgBytesPerSec=pEngine->wfx.target.Format.nAvgBytesPerSec;
#endif // ]
  double msWrittenTime=0.0;
  Ring *pRing=&pEngine->ring;
#if defined (YASAPI_RING2) // [
  WORD uFramesWritten;
#else // ] [
  DWORD dwWritten;
#endif // ]
  UINT32 uNumFramesPadding;

  if (pEngine->eState<eEngineBase)
    goto ebase;

  EnterCriticalSection(&pEngine->cs); // {

#if defined (YASAPI_TIME) // [
  switch (pEngine->eState) {
  case eEnginePauseDisconnected:
  case eEnginePauseConnected:
  case eEngineEot:
    goto exit;
  default:
    if (pEngine->eState<eEngineAudioClock)
      goto exit;

    break;
  }
#else // ] [
  if (EngineNotPlaying(pEngine))
    goto exit;

  if (pEngine->eState<eEngineRing)
    goto exit;
#endif // ]

#if defined (YASAPI_RING2) // [
  uFramesWritten=pRing->frames._uWritten;
  msWrittenTime=YASAPI_FRAMES2MS(uFramesWritten,pTarget);
#elif defined (YASAPI_RING1) // ] [
  dwWritten=RingTargetSizeWritten(pRing);
  msWrittenTime+=MulDiv(1000,dwWritten,nAvgBytesPerSec);
#endif // ]

#if defined (YASAPI_TIME) // [
  msWrittenTime-=pEngine->param.time.lpVtbl->Get(pEngine);
#else // ] [
  if (!pEngine->options.common.bLatency) {
    // Add history corrected output time.
    msWrittenTime+=EngineOutputTimeInt(pEngine);
    msWrittenTime-=pEngine->param.msHistory;
  }
#endif // ]

  if (!pEngine->options.common.bLatency&&!pEngine->param.bDoubleBuffered) {
    if (EngineNumFramesPadding(pEngine,&uNumFramesPadding)<0)
      goto exit;

    msWrittenTime+=YASAPI_FRAMES2MS(uNumFramesPadding,&pEngine->wfx.target);
  }
exit:
  LeaveCriticalSection(&pEngine->cs); // }
ebase:
  return msWrittenTime+0.5;
}

#if defined (YASAPI_RING2) // [
static void EngineCeilFrames(UINT32 *puFrames, UINT32 uChunkFrames)
{
  UINT32 uMod=*puFrames%uChunkFrames;

  if (uMod)
    *puFrames+=uChunkFrames-uMod;
}
#else // ] [
static void EngineCeil(DWORD *pdwSize, DWORD dwChunkSizeSource)
{
  DWORD dwMod=*pdwSize%dwChunkSizeSource;

  if (dwMod)
    *pdwSize+=dwChunkSizeSource-dwMod;
}
#endif // ]

#if 0 && defined (YASAPI_DEBUG) // [
// https://stackoverflow.com/questions/19738919/gcd-function-for-c
static UINT32 gcd(UINT32 a, UINT32 b)
{
  while (b) {
    a%=b;
    a^=b;
    b^=a;
    a^=b;
  }

  return a;
}

// https://stackoverflow.com/questions/19738919/gcd-function-for-c
static UINT32 lcm(UINT32 a, UINT32 b)
{
  return MulDiv(a,b,gcd(a,b));
}
#endif // ]

#if defined (YASAPI_RING2) // [
#if defined (YASAPI_RING_REALLOC) // [
//     a + b*n = c
//         b*n = c - a
static int EngineCreateRing(Engine *pEngine, DWORD dwChunkSizeSource,
    ERingAlloc eRingRealloc)
#else // ] [
static int EngineCreateRing(Engine *pEngine, DWORD dwChunkSizeSource)
#endif // ]
{
  int code=-1;
  Gtk *pGtk=&pEngine->pPlayer->gtk;
  EEngine eState=pEngine->eState;
  Ring *pRing=&pEngine->ring;
  ////////////////
#if ! defined (YASAPI_NO_OPTION_RING_SIZE) // [
  double xRingSize=pEngine->options.device.xRingSize;
#endif // ]
  WAVEFORMATEX *pSource=&pEngine->wfx.source.Format;
  UINT32 uFramesChunk=dwChunkSizeSource/pSource->nBlockAlign;
#if defined (YASAPI_NO_OPTION_RING_SIZE) // [
  double msLatency=pEngine->options.device.xMaxLatencyAssumed;
  UINT32 uFramesRing=YASAPI_MS2FRAMES(msLatency,&pEngine->wfx.target)+0.5;
#else // ] [
  UINT32 uFramesRing=xRingSize*uFramesChunk+0.5;
#endif // ]
  ////////////////
  double xPlaybackBegin=pEngine->options.device.xPlaybackBegin;
  UINT32 uFramesDevice=pEngine->param.uNumFramesDevice;
  UINT32 uFramesPlaybackBegin=xPlaybackBegin*uFramesDevice+0.5;

  /////////////////////////////////////////////////////////////////////////////
#if 0 && defined (YASAPI_RING_REALLOC) // [
  if (eRingRealloc) {
    if (pEngine->param.uFramesPlaybackBegin!=uFramesPlaybackBegin) {
      DVMESSAGE("inconsistent playback begin: %u vs. %u",
          pEngine->param.uFramesPlaybackBegin,uFramesPlaybackBegin);
      goto exit;
    }
  }
  else {
#endif // ]
    pEngine->param.uFramesPlaybackBegin=uFramesPlaybackBegin;
#if 0 && defined (YASAPI_RING_REALLOC) // [
  }
#endif // ]

DVWRITELN("uFramesRing: %u ====================================",uFramesRing);
  if (uFramesRing<uFramesPlaybackBegin)
    uFramesRing=uFramesPlaybackBegin;

DVWRITELN("uFramesRing: %u ====================================",uFramesRing);
  /////////////////////////////////////////////////////////////////////////////
  EngineCeilFrames(&uFramesRing,uFramesChunk);
DVWRITELN("uFramesRing: %u (%u %u) =============================",uFramesRing,
    uFramesRing/uFramesChunk,uFramesRing%uFramesChunk);

#if 0 && defined (YASAPI_RING_REALLOC) // [
  if (eRingRealloc) {
    if (pEngine->param.uNumFramesRing!=uFramesRing) {
      DVMESSAGE("inconsistent number of frames in ring: %u vs. %u",
          pEngine->param.uNumFramesRing,uFramesRing);
      goto exit;
    }
  }
  else {
#endif // ]
    pEngine->param.uNumFramesRing=uFramesRing;
    GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.engine.num_frames_ring);
#if 0 && defined (YASAPI_RING_REALLOC) // [
  }
#endif // ]

#if defined (YASAPI_RING_REALLOC) // [
  if (eRingRealloc) {
    if (RingReallocFrames(pRing,pEngine,uFramesRing)<0) {
      DMESSAGE("reallocating ring");
      EngineKill(pEngine,eState);
      goto exit;
    }
  }
  else {
#endif // ]
    if (RingCreateFrames(pRing,pEngine,uFramesRing)<0) {
      DMESSAGE("creating ring");
      EngineKill(pEngine,eState);
      goto exit;
    }

    ++pEngine->eState;
    DWRITELN(EngineState(pEngine));
    DASSERT(eEngineRing==pEngine->eState,exit);
#if defined (YASAPI_RING_REALLOC) // [
  }

  pEngine->param.dwChunkSizeSource=dwChunkSizeSource;
#endif // ]

#if defined (YASAPI_CONTROL_START) // [
  pEngine->monitor.xStart=(double)uFramesPlaybackBegin/uFramesRing;
  GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.start);
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  code=0;
exit:
  return code;
}
#else // ] [
static int EngineCreateRing(Engine *pEngine, DWORD dwChunkSizeSource)
{
  int code=-1;
  Ring *pRing=&pEngine->ring;
  Gtk *pGtk=&pEngine->pPlayer->gtk;
  WAVEFORMATEX *pSource=&pEngine->wfx.source.Format;
  EEngine eState=pEngine->eState;
  double xPlaybackBegin=pEngine->options.device.xPlaybackBegin;
  UINT32 uNumFramesDevice=pEngine->param.uNumFramesDevice;
#if ! defined (YASAPI_RING1) && ! defined (YASAPI_RING2) // [
  WAVEFORMATEXTENSIBLE *pWfxTarget=&pEngine->wfx.target;
  WAVEFORMATEXTENSIBLE *pWfxSource=&pEngine->wfx.source;
#endif // ]
  DWORD dwSizeSharedSource;
  DWORD dwPlaybackBeginSource;
  DWORD dwSizeRingSource;
#if ! defined (YASAPI_RING1) && ! defined (YASAPI_RING2) // [
  IRingVtbl *lpRingVtbl;
  int bFormatTransform;
#endif // ]
#if 0 && defined (YASAPI_DEBUG) // [
  UINT32 uChunkSize=dwChunkSizeSource/nBlockAlign;
  UINT32 uLcm=lcm(pEngine->param.uNumFramesDevice,uChunkSize);

  DVWRITELN("uNumFramesDevice=%u uChunkSize=%u uLcm=%u %u %u",
      pEngine->param.uNumFramesDevice,uChunkSize,uLcm,
          uLcm/pEngine->param.uNumFramesDevice,uLcm/uChunkSize);
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  // We round up dwSizeRing/dwPlaybackBegin to a multiple of len in order that
  // multiple chunks fit.
  dwSizeSharedSource=uNumFramesDevice*pSource->nBlockAlign;

  dwPlaybackBeginSource=xPlaybackBegin*dwSizeSharedSource+0.5;
  EngineCeil(&dwPlaybackBeginSource,dwChunkSizeSource);
  pEngine->param.source.dwPlaybackBegin=dwPlaybackBeginSource;

  dwSizeRingSource=pEngine->options.device.xRingSize*dwChunkSizeSource+0.5;

  // We make sure that at least two chunks fit into the buffer.
  if (dwSizeRingSource<dwPlaybackBeginSource+(dwChunkSizeSource<<1))
    dwSizeRingSource=dwPlaybackBeginSource+(dwChunkSizeSource<<1);

  pEngine->param.uNumFramesRing=dwSizeRingSource/pSource->nBlockAlign;
  DVWRITELN("dwPlaybackBegin=%lu dwSizeRing=%lu",
      dwPlaybackBeginSource/pSource->nBlockAlign,
      dwSizeRingSource/pSource->nBlockAlign);

  GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.engine.num_frames_ring);

  /////////////////////////////////////////////////////////////////////////////
  // The ring buffer is set up wrt to the target format.
#if ! defined (YASAPI_RING1) && ! defined (YASAPI_RING2) // [
  bFormatTransform=memcmp(&pWfxTarget->Format,&pWfxSource->Format,
      sizeof pWfxTarget->Format);

  if (eVolumeSoftware==pEngine->options.common.eVolume||bFormatTransform)
    lpRingVtbl=&gcIRingTransformVtbl;
  else
    lpRingVtbl=&gcIRingPlainVtbl;

  if (RingCreate(pRing,lpRingVtbl,pEngine,dwSizeRingSource)<0) {
    DMESSAGE("creating ring");
    EngineKill(pEngine,eState);
    goto exit;
  }
#else // ] [
  if (RingCreate(pRing,pEngine,dwSizeRingSource)<0) {
    DMESSAGE("creating ring");
    EngineKill(pEngine,eState);
    goto exit;
  }
#endif // ]

  ++pEngine->eState;
  DWRITELN(EngineState(pEngine));
  DASSERT(eEngineRing==pEngine->eState,exit);
  code=0;
exit:
  DWRITELN(EngineState(pEngine));
  return code;
}
#endif // ]

#if defined (YASAPI_RING2) // [
#if defined (YASAPI_RING_REALLOC) // [
static int EngineChunkSizeKnown(Engine *pEngine, DWORD dwChunkSizeSource)
{
   return pEngine->param.dwChunkSizeSource==dwChunkSizeSource;
}
#endif // ]

#if defined (YASAPI_API) // [
static int _EngineWrite(Engine *pEngine, const BYTE *buf,
    DWORD dwChunkSizeSource, int bLocked)
#else // ] [
int EngineWrite(Engine *pEngine, BYTE *buf, DWORD dwChunkSizeSource)
#endif // ]
{
  int code=-1;
  Player *pPlayer=pEngine->pPlayer;
  Gtk *pGtk=&pPlayer->gtk;
  Ring *pRing=&pEngine->ring;
  WORD uFramesPlaybackBegin;
  WORD uFramesWritten;
  WORD uFramesMax;
#if defined (YASAPI_RING_REALLOC) // [
  DWORD dwSourceSize;
#endif // ]
  UINT32 uMsLatency;

  if (pEngine->eState<eEngineBase) {
    DMESSAGE("engine not initialized");
    goto ebase;
  }

#if defined (YASAPI_API) // [
  if (!bLocked)
#endif // ]
    EnterCriticalSection(&pEngine->cs); // {

#if defined (YASAPI_RING_REALLOC) // [
  switch (pEngine->eState) {
  case eEngineBase:
    // During "open" we thought this is "gapless". Meanwhile the sesssion was
    // terminated.
    if (EngineCreateGaplessSession(pEngine)<0) {
      DMESSAGE("creating gapless sequence");
      goto exit;
    }

    ++pEngine->pPlayer->stats.uNumGaplessBroken;
    GtkUpdateWidget(pGtk,
        &pGtk->config.config.monitor.player.num_gapless_broken);

#if 0 // [
    if (EngineCreateRing(pEngine,dwChunkSize,eRingCreate)<0) {
      DMESSAGE("creating ring");
      goto exit;
    }
#endif // ]

    break;
  case eEngineDraining:
  case eEngineDrainingFinal:
    // We're still in state "eEngineDraining" ...
    EngineKill(pEngine,eEngineTimerStarted);
    ++pEngine->stats.uNumTracks;
    ++pEngine->pPlayer->stats.uNumTracks;

    if (EngineCreateRing(pEngine,dwChunkSizeSource,eRingRealloc)<0) {
      DMESSAGE("creating ring");
      goto exit;
    }

    GtkUpdateDialog(&pPlayer->gtk,&pPlayer->gtk.config);
    break;
  default:
    if (pEngine->eState<eEngineRing) {
      if (EngineCreateRing(pEngine,dwChunkSizeSource,eRingCreate)<0) {
        DMESSAGE("creating ring");
        goto exit;
      }
    }
    else if (!EngineChunkSizeKnown(pEngine,dwChunkSizeSource)) {
      // This is special to "in_mp3" because it has not a constant block size.
      dwSourceSize=RingSourceSizeAvailable(pRing)+dwChunkSizeSource;

      if (RingSourceSize(pRing)<dwSourceSize
          &&EngineCreateRing(pEngine,dwChunkSizeSource,eRingRealloc)<0) {
        DMESSAGE("creating ring");
        goto exit;
      }
    }

    break;
  }
#elif defined (YASAPI_OPTION_GAPLESS) // [
  /////////////////////////////////////////////////////////////////////////////
  if (eEngineBase==pEngine->eState) {
    // During "open" we thought this is "gapless". Meanwhile the sesssion was
    // terminated.
    if (EngineCreateGaplessSession(pEngine)<0) {
      DMESSAGE("creating gapless sequence");
      goto exit;
    }

    ++pEngine->pPlayer->stats.uNumGaplessBroken;
    GtkUpdateWidget(pGtk,
        &pGtk->config.config.monitor.player.num_gapless_broken);
  }
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  switch (pEngine->eState) {
  case eEnginePauseDisconnected:
  case eEnginePauseConnected:
    DWRITELN("Warning: pause");
    goto exit;
  default:
    break;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (pEngine->eState<eEngineRing) {
    DASSERT_ENGINE_STATE(pEngine,eEngineAudioClientInititial,exit);

#if defined (YASAPI_RING_REALLOC) // [
    if (EngineCreateRing(pEngine,dwChunkSizeSource,eRingCreate)<0) {
      DMESSAGE("creating ring");
      goto exit;
    }
#else // ] [
    if (EngineCreateRing(pEngine,dwChunkSize)<0) {
      DMESSAGE("creating ring");
      goto exit;
    }
#endif // ]
  }

  uFramesPlaybackBegin=pEngine->param.uFramesPlaybackBegin;

  if (RingWrite(pRing,buf,dwChunkSizeSource)<0) {
    DMESSAGE("writing to ring");
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  uFramesWritten=pRing->frames._uWritten;

  if (pEngine->stats.uNumFramesMax<uFramesWritten) {
    pEngine->stats.uNumFramesMax=uFramesWritten;
    GtkUpdateWidget(pGtk,
        &pGtk->config.config.monitor.engine.num_frames_latency);
  }

#if 1 // [
  /////////////////////////////////////////////////////////////////////////////
  uMsLatency=YASAPI_FRAMES2MS(uFramesWritten,&pEngine->wfx.source)+0.5;

  if (pPlayer->stats.uMsLatencyMax<uMsLatency) {
    pPlayer->stats.uMsLatencyMax=uMsLatency;
    GtkUpdateWidget(pGtk,
        &pGtk->config.config.monitor.player.ms_latency);
  }
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  if (eEngineStarted<=pEngine->eState)
    goto success;

  if (uFramesWritten<uFramesPlaybackBegin) {
    ///////////////////////////////////////////////////////////////////////////
    uMsLatency=YASAPI_FRAMES2MS(uFramesWritten,&pEngine->wfx.source)+0.5;

#if 0 // [
#if 0 // [
    if (pPlayer->stats.uMsLatencyMax<uMsLatency) {
#endif // ]
    pPlayer->stats.uMsLatencyMax=uMsLatency;
    GtkUpdateWidget(pGtk,
        &pGtk->config.config.monitor.player.ms_latency);
#if 0 // [
    }
#endif // ]
#endif // ]

    DVWRITELN("*** underflow: %u (%u %u)",uFramesWritten,uFramesPlaybackBegin,
        dwChunkSizeSource);
    //uFramesMax=pRing->lpVtbl->FramesMaximum(pRing);
    uFramesMax=RingFramesMaximum(pRing);
    pEngine->monitor.xWritten=(double)uFramesWritten/uFramesMax;
    GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.written);
    goto success;
  }

  DVWRITELN("*** starting: %u (%u)",uFramesWritten,uFramesPlaybackBegin);
  EngineSend(pEngine,eCommandStart,eEngineNull,0);
  DWRITELN(EngineState(pEngine));
success:
  code=0;
exit:
#if defined (YASAPI_API) // [
  if (!bLocked)
#endif // ]
    LeaveCriticalSection(&pEngine->cs); // }
ebase:
  return code;
}
#else // ] [
#if defined (YASAPI_API) // [
static int _EngineWrite(Engine *pEngine, const BYTE *buf, DWORD dwChunkSize,
    int bLocked)
#else // ] [
int EngineWrite(Engine *pEngine, BYTE *buf, DWORD dwChunkSize)
#endif // ]
{
  int code=-1;
  Ring *pRing=&pEngine->ring;
  Player *pPlayer=pEngine->pPlayer;
  Gtk *pGtk=&pPlayer->gtk;
  DWORD dwSizeSource;
  DWORD dwWrittenSource;
  DWORD dwWrittenTarget;
  UINT32 uNumFrames;
  UINT32 uMsLatency;

  if (pEngine->eState<eEngineBase) {
    DMESSAGE("engine not initialized");
    goto ebase;
  }

#if defined (YASAPI_API) // [
  if (!bLocked)
#endif // ]
    EnterCriticalSection(&pEngine->cs); // {

  switch (pEngine->eState) {
  case eEnginePauseDisconnected:
  case eEnginePauseConnected:
    goto exit;
  default:
    break;
  }

  if (pEngine->eState<eEngineRing&&EngineCreateRing(pEngine,dwChunkSize)<0) {
    DMESSAGE("creating ring");
    goto exit;
  }

#if 0 // [
DVWRITELN("dwChunkSize=%lu {",dwChunkSize);
  dwChunkSize=pRing->lpVtbl->target.ConvertSize(pRing,dwChunkSize);
DVWRITELN("dwChunkSize=%lu }",dwChunkSize);
#endif // ]

  if (RingWrite(&pEngine->ring,buf,dwChunkSize)<0) {
    DMESSAGE("writing to ring");
    goto exit;
  }

#if defined (YASAPI_RING1) // [
  dwSizeSource=RingSourceSize(pRing);
  dwWrittenSource=RingSourceSizeWritten(pRing);
  dwWrittenTarget=RingTargetSizeWritten(pRing);
#else // ] [
  dwSizeSource=pRing->lpVtbl->source.Size(pRing);
  dwWrittenSource=pRing->lpVtbl->source.Written(pRing);
  dwWrittenTarget=pRing->lpVtbl->target.Written(pRing);
#endif // ]
  uNumFrames=YASAPI_SIZE2FRAMES(dwWrittenTarget,&pEngine->wfx.target);
  uMsLatency=YASAPI_FRAMES2MS(uNumFrames,&pEngine->wfx.target)+0.5;

  if (pPlayer->stats.uMsLatencyMax<uMsLatency) {
    pPlayer->stats.uMsLatencyMax=uMsLatency;
    GtkUpdateWidget(pGtk,
        &pGtk->config.config.monitor.player.ms_latency);
  }

  if (pEngine->stats.uNumFramesMax<uNumFrames) {
    pEngine->stats.uNumFramesMax=uNumFrames;
    GtkUpdateWidget(pGtk,
        &pGtk->config.config.monitor.engine.num_frames_latency);
  }

  if (eEngineStarted<=pEngine->eState)
    goto success;

  if (dwWrittenSource<pEngine->param.source.dwPlaybackBegin) {
    DVWRITELN("*** undeflow: %lu (%lu)",
        dwWrittenSource/pEngine->wfx.source.Format.nBlockAlign,
        pEngine->param.source.dwPlaybackBegin
            /pEngine->wfx.source.Format.nBlockAlign);
    pEngine->monitor.xWritten=(double)dwWrittenSource/dwSizeSource;
    GtkUpdateWidget(pGtk,&pGtk->config.config.monitor.written);
    goto success;
  }

  DVWRITELN("*** starting: %lu (%lu)",
      dwWrittenSource/pEngine->wfx.source.Format.nBlockAlign,
      pEngine->param.source.dwPlaybackBegin
          /pEngine->wfx.source.Format.nBlockAlign);

  EngineSend(pEngine,eCommandStart,eEngineNull,0);
  DWRITELN(EngineState(pEngine));
success:
  code=0;
exit:
#if defined (YASAPI_API) // [
  if (!bLocked)
#endif // ]
    LeaveCriticalSection(&pEngine->cs); // }
ebase:
  return code;
}
#endif // ]

#if defined (YASAPI_API) // [
int EngineWrite(Engine *pEngine, BYTE *buf, DWORD dwChunkSize)
{
  return _EngineWrite(pEngine,buf,dwChunkSize,0);
}
#endif // ]

void EngineVolume(Engine *pEngine, Player *pPlayer)
{
  if (pEngine->eState<eEngineBase&&EngineCreateBase(pEngine,pPlayer)<0) {
    DWRITELN("engine not initialized");
    goto ebase;
  }

  EnterCriticalSection(&pEngine->cs); // {

  if (pEngine->eState<eEngineStarted) {
    DWRITELN(EngineState(pEngine));
    goto exit;
  }

  pEngine->param.pIEngineVolumeVtbl->Set(pEngine);
exit:
  LeaveCriticalSection(&pEngine->cs); // }
ebase:
  return;
}

#if defined (YASAPI_OPTION_GAPLESS) // [
void EngineSetActive(Engine *pEngine, int bActive)
{
  if (eEngineBase<pEngine->eState) {
    EnterCriticalSection(&pEngine->cs); // {
    pEngine->param.bActive=bActive;
    DVWRITELN("==> bActive: %d <==================================",bActive);

    if (!bActive&&eEngineEot==pEngine->eState)
        EngineKill(pEngine,eEngineBase);

    LeaveCriticalSection(&pEngine->cs); // }
  }
}
#endif // ]

#if defined (YASAPI_API) // [
int EngineApiWrite(Engine *pEngine, int msTimeout, const char *buf,
    DWORD dwChunkSize)
{
  int code=-1;

  if (pEngine->eState<eEngineBase)
    goto enoinit;

  EnterCriticalSection(&pEngine->cs); // {

  pEngine->param.bApiKilled=0;

  for (;;) {
    if (pEngine->param.bApiKilled)
      goto exit;
    else if(dwChunkSize<=EngineAvailable(pEngine)) {
      code=_EngineWrite(pEngine,(const BYTE *)buf,dwChunkSize,1);
      goto exit;
    }

    ResetEvent(pEngine->thread.hApi);
    LeaveCriticalSection(&pEngine->cs); // {
    WaitForSingleObject(pEngine->thread.hApi,msTimeout);
    EnterCriticalSection(&pEngine->cs); // }
  }
exit:
  LeaveCriticalSection(&pEngine->cs); // }
enoinit:
  return code;
}

int EngineApiEndOfTrack(Engine *pEngine, int msTimeout)
{
  int code=-1;

  if (pEngine->eState<eEngineBase)
    goto enoinit;

  EnterCriticalSection(&pEngine->cs); // {

#if defined (YASAPI_OPTION_GAPLESS) // [
  if (pEngine->options.common.bGapless) {
    ++pEngine->eState;
    DASSERT_ENGINE_STATE(pEngine,eEngineDraining,exit);
    code=0;
    goto exit;
  }
#endif // ]

  pEngine->param.bApiKilled=0;

  for (;;) {
    if (pEngine->param.bApiKilled)
      goto exit;
    else if (!EngineIsPlaying(pEngine)) {
      code=0;
      goto exit;
    }

    LeaveCriticalSection(&pEngine->cs); // {
    WaitForSingleObject(pEngine->thread.hApi,msTimeout);
    EnterCriticalSection(&pEngine->cs); // }
  }
exit:
#if 0 // [
  DVWRITELN("=== target: %u/%u/%u ===",
      pEngine->wfx.target.Format.nSamplesPerSec,
      pEngine->wfx.target.Format.nChannels,
      pEngine->wfx.target.Format.wBitsPerSample);
#endif // ]
  LeaveCriticalSection(&pEngine->cs); // }
enoinit:
  return code;
}

void EngineApiKill(Engine *pEngine)
{
  if (pEngine->eState<eEngineBase)
    goto enoinit;

  EnterCriticalSection(&pEngine->cs); // {
  pEngine->param.bApiKilled=1;
  SetEvent(pEngine->thread.hApi);
  LeaveCriticalSection(&pEngine->cs); // }
enoinit: ;
}
#endif // ]

///////////////////////////////////////////////////////////////////////////////
void EngineSend(Engine *pEngine, const ECommand eCommand, EEngine eArgument,
  int bLock)
{
  Command command={ eCommand, eArgument };

  if (bLock)
    EnterCriticalSection(&pEngine->cs); // {

  pEngine->thread.pCommand=&command;
  SetEvent(pEngine->thread.hProducer);

  while (pEngine->thread.pCommand) {
    LeaveCriticalSection(&pEngine->cs); // {
    WaitForSingleObject(pEngine->thread.hConsumer,INFINITE);
    EnterCriticalSection(&pEngine->cs); // }
  }

  if (bLock)
    LeaveCriticalSection(&pEngine->cs); // }
}

///////////////////////////////////////////////////////////////////////////////
const char *EngineStateEx(EEngine eState)
{
  const char *mnemonic[1+eEngineMax]={
    "undefined",
    "eEngineNull",
    "eEngineCriticalSection",
    "eEngineTimerEvent",
    "eEngineTimer",
    "eEngineEventProducer",
    "eEngineEventConsumer",
#if defined (YASAPI_OPTION_GAPLESS) // [
    "eEngineEventEot",
#endif // ]
#if defined (YASAPI_API) // [
    "eEngineEventApi",
#endif // ]
    "eEngineThread",
    "eEngineThreadRunning",
#if defined (YASAPI_ENGINE_ENUMERATOR) // [
    "eEngineDeviceEnumerator",
    "eEngineDevice",
#endif // ]
#if defined (YASAPI_OPTION_GAPLESS) // [
    "eEngineClose",
#endif // ]
    "eEngineAudioClientInititial",
    "eEngineRing",
    "eEnginePauseDisconnected",
    "eEngineAudioClient",
    "eEngineVolume",
    "eEngineAudioClock",
    "eEngineAudioRenderClient",
    "eEnginePauseConnected",
    "eEngineAudioClientStarted",
    "eEngineTimerStarted",
    "eEngineDraining",
    "eEngineDrainingFinal",
    "eEngineEot",
  };

  return mnemonic[0<=eState&&eState<eEngineMax?1+eState:0];
}

const char *EngineState(Engine *pEngine)
{
  return EngineStateEx(pEngine->eState);
}

#if defined (YASAPI_DEBUG) // [
const char *EngineCommandEx(ECommand eCommand)
{
  const char *mnemonic[1+eCommandMax]={
    "undefined",
    "eCommandNull",
    "eCommandExit",
    "eCommandStart",
    "eCommandFlush",
    "eCommandKill",
  };

  return mnemonic[0<=eCommand&&eCommand<eCommandMax?1+eCommand:0];
}

const char *EngineCommand(Engine *pEngine)
{
  Command *pCommand=pEngine->thread.pCommand;

  return pCommand?EngineCommandEx(pCommand->eCommand):"NULL";
}
#endif // ]
