/*
 * yasapi_trace.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

#if defined (YASAPI_DEBUG) // [
///////////////////////////////////////////////////////////////////////////////
static CRITICAL_SECTION cs;
static int bCs;
static int bConsole;

///////////////////////////////////////////////////////////////////////////////
int TraceCreate(Out_Module *pOut)
{
  InitializeCriticalSection(&cs);
  bCs=1;

  if (pOut) {
    if (0!=(bConsole=AllocConsole())) {
      freopen("conin$", "r", stdin);
      freopen("conout$", "w", stdout);
      freopen("conout$", "w", stderr);
      DWRITELN("Console allocated.");
    }
  }
  else
    bConsole=0;

  return 0;
}

void TraceDestroy(void)
{
  fputs("Hit some key ...",stderr);
  fgetc(stdin);

  if (bConsole)
    FreeConsole();

  if (bCs) {
    bCs=0;
    DeleteCriticalSection(&cs);
  }
}

///////////////////////////////////////////////////////////////////////////////
static void TraceLock(void)
{
  if (bCs)
    EnterCriticalSection(&cs);
}

static void TraceUnlock(void)
{
  if (bCs)
    LeaveCriticalSection(&cs);
}

///////////////////////////////////////////////////////////////////////////////
void TraceMarkLn(const char *pFile, int nLine, const char *pFunc)
{
  TraceLock(); // {
  fprintf(stderr,"[%s:%d:%s]\n",yasapi_basenamea(pFile),nLine,pFunc);
  fflush (stderr);
  TraceUnlock(); // }
}

void TraceWriteLn(const char *pFile, int nLine, const char *pFunc,
    const char *pLine)
{
  TraceLock(); // {
  fprintf(stderr,"[%s:%d:%s] ",yasapi_basenamea(pFile),nLine,pFunc);
  fputs(pLine,stderr);
  fputs("\n",stderr);
  fflush(stderr);
  TraceUnlock(); // }
}
void TraceWriteLnW(const char *pFile, int nLine, const char *pFunc,
    const wchar_t *pLine)
{
  TraceLock(); // {
  fprintf(stderr,"[%s:%d:%s] ",yasapi_basenamea(pFile),nLine,pFunc);
  fputws(pLine,stderr);
  fputc('\n',stderr);
  fflush(stderr);
  TraceUnlock(); // }
}

void TraceVWriteLn(const char *pFile, int nLine, const char *pFunc,
    const char *pFormat, ...)
{
  va_list ap;

  TraceLock(); // {
  fprintf(stderr,"[%s:%d:%s] ",yasapi_basenamea(pFile),nLine,pFunc);
  fflush (stderr);
  va_start(ap,pFormat);
  vfprintf(stderr,pFormat,ap);
  va_end(ap);
  fputwc(L'\n',stderr);
  fflush (stderr);
  TraceUnlock(); // }
}

void TraceVWriteLnW(const char *pFile, int nLine, const char *pFunc,
    const wchar_t *pFormat, ...)
{
  va_list ap;

  TraceLock(); // {
  fprintf(stderr,"[%s:%d:%s] ",yasapi_basenamea(pFile),nLine,pFunc);
  va_start(ap,pFormat);
  vfwprintf(stderr,pFormat,ap);
  va_end(ap);
  fputs("\n",stderr);
  fflush (stderr);
  TraceUnlock(); // }
}
#endif // ]
